/*************************************************************
 * $Id: hash.c 342 2003-10-13 16:46:18Z clerger $
 * Copyright (C) 1997, 2003 Eric de la Clergerie
 * ------------------------------------------------------------
 *
 *    Hash -- Hash Table management  
 *
 * ------------------------------------------------------------
 * Description
 *    This file comes from the wamcc distribution
 *    (see following header)
 * ------------------------------------------------------------
 */

/*-------------------------------------------------------------------------*/
/* C Run-Time                                           Daniel Diaz - 1996 */
/* Hash table management                                                   */
/*                                                                         */
/* hash.c                                 Copyright (C) 1996, INRIA France */
/*-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>

#include "bigloo.h"

#include "hash.h"
#include "fol.h"

/*---------------------------------*/
/* Constants                       */
/*---------------------------------*/

static const struct {
    unsigned long power;
    unsigned long prime;
} list[] = {
    {         4,             5},            /* 2^2 */
    {         8,            11},            /* 2^3 */
    {        16,            17},            /* 2^4 */
    {        32,            37},            /* 2^5 */
    {        64,            67},            /* 2^6 */
    {       128,           131},            /* 2^7 */
    {       256,           257},            /* 2^8 */
    {       512,           521},            /* 2^9 */
    {      1024,          1031},            /* 2^10 */
    {      2048,          2053},            /* 2^11 */
    {      4096,          4099},            /* 2^12 */
    {      8192,          8191},            /* 2^13 */
    {     16384,         16381},            /* 2^14 */
    {     32768,         32771},            /* 2^15 */
    {     65536,         65537},            /* 2^16 */
    {    131072,        131071},            /* 2^17 */
    {    262144,        262147},            /* 2^18 */
    {0,             0}
};

/*---------------------------------*/
/* Type Definitions                */
/*---------------------------------*/
 
typedef struct hash_node *HashNode;

struct hash_node {
    HashNode next;              /* hash node are chained in a bucket */
    HashNode linear;            /* hash node are chained fora linear traversal
                                 * Hash_Next */
    void *   linear_prev;       /* to update linear in case of delete */
    long key;                   /* the hash key */
        /* the rest of the elem comes here */
};

typedef struct Hash_Table
{
    unsigned long size;
    unsigned long elem_size;
    unsigned long nb_elem;
    HashNode  first;
    HashNode  *table;         /* ptr to the table, indirection to allow reallocation */
} *HashTable;

/*---------------------------------*/
/* Global Variables                */
/*---------------------------------*/

/*---------------------------------*/
/* Function Prototypes             */
/*---------------------------------*/

static 
HashNode *Hash_Locate           (const char *t,const long key);

#define HASH_STATIC_SIZE           sizeof(struct Hash_Table)

#define Tbl_Size(t)                (((HashTable)t)->size)
#define Elem_Size(t)               (((HashTable)t)->elem_size)
#define Nb_Elem(t)                 (((HashTable)t)->nb_elem)
#define Hsh_Table(t)               ((((HashTable)t)->table))
#define Tbl_First(t)               (((HashTable)t)->first)

#define Hash_Function(k,size)      ((unsigned long)(k) % (size))

/*-------------------------------------------------------------------------*/
/* A hash table consists of a header (tbl_size, elem_size, nb_elem) and a  */
/* table of tbl_size pointers to nodes.                                    */
/* Each node records a pointer to the next node, and a user element whose  */
/* size is elem_size. Each element must begin with the key (a long).       */
/*-------------------------------------------------------------------------*/

inline static
unsigned long
local_hash_fol(const fol_t a) 
{
    unsigned long n;
    if (FOLSMBP(a)) {
        n=FOLSMB_INDEX(a);
    } else if (FOLCMPP(a)) {
        n=((unsigned long)a) >> 4;      
    } else {
        n=((unsigned long)a) >> 2;      
    }
    return n;
}

static unsigned long
Hash_Find_Right_Size(unsigned long size) 
{
/*
  int i=0;
    for(;;i++) {
        if (list[i].power == 0 || list[i].power > size) {
            return (list[i-1].prime);
        }
    }
*/
    return size;
}

/*-------------------------------------------------------------------------*/
/* HASH_ALLOC_TABLE                                                        */
/*                                                                         */
/*-------------------------------------------------------------------------*/
char *Hash_Alloc_Table(const int tbl_size, const int elem_size)

{
        /* at least try 4 * buckets */
    unsigned long size = Hash_Find_Right_Size((tbl_size < 4) ? 4 : tbl_size); 
    HashNode *table = (HashNode *)GC_MALLOC( size*sizeof(HashNode));
    HashTable tbl= (HashTable) GC_MALLOC( HASH_STATIC_SIZE );
    if (tbl==NULL)
        return NULL;
    
    Tbl_Size(tbl) = size;
    Elem_Size(tbl)= elem_size;
    Nb_Elem(tbl)=0;
    Hsh_Table(tbl) = table;
    Tbl_First(tbl)=NULL;
    return (char *)tbl;
}

/*-------------------------------------------------------------------------*/
/* HASH_REALLOC_TABLE                                                      */
/*                                                                         */
/*-------------------------------------------------------------------------*/

static void
Hash_Realloc_Table(const char *tbl)
{
    int       tbl_size=Tbl_Size(tbl);
    HashNode *t=Hsh_Table(tbl);
    HashNode *endt=t+tbl_size;
    HashNode  p,p1;
    HashNode *prev;
    unsigned long new_tbl_size = Hash_Find_Right_Size(tbl_size << 1);
    HashNode *new_t = (HashNode *)GC_MALLOC(new_tbl_size*sizeof(HashNode));

    if (new_t==NULL) {
        return;                 /* error */
    }

    Hsh_Table(tbl) = new_t;
    Tbl_Size(tbl) = new_tbl_size;

    while (t<endt) {
        p=*t++;
        while(p){
            prev=Hash_Locate(tbl,p->key); 
            p1=p;
            p=p->next;
            *prev=p1;
            p1->next=NULL;
        }
    }
}

/*-------------------------------------------------------------------------*/
/* HASH_FREE_TABLE                                                         */
/*                                                                         */
/*-------------------------------------------------------------------------*/
void Hash_Free_Table(char *tbl)

{
 int       tbl_size=Tbl_Size(tbl);
 HashNode *t=Hsh_Table(tbl);
 HashNode *endt=t+tbl_size;
 HashNode  p,p1;

 while(t<endt) {
     p=*t++;
     while(p){
         p1=p;
         p=p->next;
//         free(p1);
     }
 }
}

/*-------------------------------------------------------------------------*/
/* HASH_DELETE_ALL                                                         */
/*                                                                         */
/*-------------------------------------------------------------------------*/
void
Hash_Delete_All(const char *tbl)
{
    int       tbl_size=Tbl_Size(tbl);
    HashNode *t=Hsh_Table(tbl);
    HashNode *endt=t+tbl_size;

    while(t<endt) {
        *t++=NULL;
    }
    Nb_Elem(tbl)=0;
}

/*-------------------------------------------------------------------------*/
/* HASH_INSERT                                                             */
/*                                                                         */
/*-------------------------------------------------------------------------*/
char *
Hash_Insert(const char *tbl,const char *elem,const int replace)
{
    int       elem_size=Elem_Size(tbl);
    long      key= *(long *) elem;
    HashNode *prev;
    HashNode  p;

    prev=Hash_Locate(tbl,key);
    p=*prev;
    
    if (p==NULL)                                    /* the key does not exist */
    {
        p=(HashNode) GC_MALLOC(sizeof(struct hash_node)-sizeof(long)+elem_size);
        if (p==NULL)
            return NULL;
        
        p->next=NULL;
        p->linear=Tbl_First(tbl);
        if (Tbl_First(tbl))
            Tbl_First(tbl)->linear_prev=(void *)&(p->linear);
        Tbl_First(tbl)=p;
        p->linear_prev=(void *)&(Tbl_First(tbl));
        *prev=p;
        Nb_Elem(tbl)++;
        memcpy((char *) (&(p->key)),elem,elem_size);
        if (Nb_Elem(tbl)>Tbl_Size(tbl)) { /* realloc */
            Hash_Realloc_Table(tbl);
        }
    }
    else if (replace) {
        memcpy((char *) (&(p->key)),elem,elem_size);
    }
 
    return (char *) (&(p->key));
}

/*-------------------------------------------------------------------------*/
/* HASH_FIND                                                               */
/*                                                                         */
/*-------------------------------------------------------------------------*/
char *
Hash_Find(const char *tbl, const long key)
{
 HashNode *prev;
 HashNode  p;

 prev=Hash_Locate(tbl,key);
 p=*prev;

 if (p==NULL)                                    /* the key does not exist */
     return NULL;

 return (char *) (&(p->key));
}

/*-------------------------------------------------------------------------*/
/* HASH_DELETE                                                             */
/*                                                                         */
/*-------------------------------------------------------------------------*/
void
Hash_Delete(const char *tbl, const long key)
{
    HashNode *prev;
    HashNode  p;
    
    prev=Hash_Locate(tbl,key);
    p=*prev;
    
    if (p==NULL)                                    /* the key does not exist */
        return;
    
    *prev=p->next;
    if (p->linear)
        p->linear->linear_prev=p->linear_prev;
    *(HashNode *)(p->linear_prev)=p->linear;
    Nb_Elem(tbl)--;
}

/*-------------------------------------------------------------------------*/
/* HASH_LOCATE                                                             */
/*                                                                         */
/* This function returns the address of the pointer to the node associated */
/* to the key (if the pointer is NULL the key is not in the table).        */
/*-------------------------------------------------------------------------*/
static
HashNode *Hash_Locate(const char *tbl,const long key)
{
    HashNode *t = Hsh_Table(tbl)+Hash_Function(local_hash_fol((fol_t)key),
                                               Tbl_Size(tbl));
    HashNode p;
 
    for(p=*t;p;p=p->next) {
        if (p->key==key)
            break;     
        t=&p->next;
    }

    return t;
}

/*-------------------------------------------------------------------------*/
/* HASH_FIRST                                                              */
/*                                                                         */
/* Hash_First and Hash_Next make it possible to scan a hash table.         */
/* Example of use:                                                         */
/*                                                                         */
/*   HashScan scan;                                                        */
/*   char     *buff_ptr;                                                   */
/*                                                                         */
/*   for(buff_ptr=Hash_First(tbl,&scan);buff_ptr;buff_ptr=Hash_Next(&scan))*/
/*       Display_Element(buff_ptr);                                        */
/*-------------------------------------------------------------------------*/
char *
Hash_First(const char *tbl, HashScan *scan)
{
    scan->current=(char *) (Tbl_First(tbl));
    return Hash_Next(scan);
}

/*-------------------------------------------------------------------------*/
/* HASH_NEXT                                                               */
/*                                                                         */
/*-------------------------------------------------------------------------*/
char *
Hash_Next(HashScan *scan)
{
    HashNode  p;
    
    p=(HashNode) (scan->current);

    if (p==NULL)
        return NULL;

    scan->current=(char *) (p->linear);
    return (char *) (&(p->key));
}

/*-------------------------------------------------------------------------*/
/* HASH_TABLE_SIZE                                                         */
/*                                                                         */
/*-------------------------------------------------------------------------*/
int Hash_Table_Size(const char *tbl)
{
    return Tbl_Size(tbl);
}

/*-------------------------------------------------------------------------*/
/* HASH_NB_ELEMENTS                                                        */
/*                                                                         */
/*-------------------------------------------------------------------------*/
int Hash_Nb_Elements(const char *tbl)
{
    return Nb_Elem(tbl);
}

